@@ -1,4 +1,4 @@
-# pylint: disable=too-many-public-methods, missing-docstring, too-many-lines, use-of-eval, disallowed-variable, no-self-use
+# pylint: disable=too-many-public-methods, missing-docstring, eval-used, too-many-lines, no-self-use, disallowed-name, unspecified-encoding
 
 """
     Unit tests for simpleeval.
@@ -24,6 +24,7 @@ from simpleeval import (
     FunctionNotDefined,
     InvalidExpression,
     NameNotDefined,
+    OperatorNotDefined,
     SimpleEval,
     simple_eval,
 )
@@ -82,6 +83,13 @@ class TestBasic(DRYTest):
             -10,
         )
 
+    def test_bit_ops(self):
+        self.t("62 ^ 20", 42)
+        self.t("62 ^ 100", 90)
+        self.t("8 | 34", 42)
+        self.t("100 & 63", 36)
+        self.t("~ -43", 42)
+
     def test_not(self):
         self.t("not False", True)
         self.t("not True", False)
@@ -109,6 +117,7 @@ class TestBasic(DRYTest):
         self.t("1 < 2 < 3 < 4", 1 < 2 < 3 < 4)
         self.t("1 < 2 > 3 < 4", 1 < 2 > 3 < 4)
 
+        # pylint: disable=comparison-with-itself
         self.t("1<2<1+1", 1 < 2 < 1 + 1)
         self.t("1 == 1 == 2", 1 == 1 == 2)
         self.t("1 == 1 < 2", 1 == 1 < 2)
@@ -189,6 +198,49 @@ class TestBasic(DRYTest):
         with self.assertRaises(FeatureNotAvailable):
             self.t("{22}", False)
 
+    def test_empty_string_not_allowed(self):
+        with self.assertRaises(InvalidExpression):
+            self.t("", False)
+
+
+class TestEvaluator(DRYTest):
+    """Tests for how the SimpleEval class does things"""
+
+    def test_only_evalutate_first_statement(self):
+        # it only evaluates the first statement:
+        with warnings.catch_warnings(record=True) as ws:
+            self.t("11; x = 21; x + x", 11)
+        self.assertIsInstance(ws[0].message, simpleeval.MultipleExpressions)
+
+    def test_parse_and_use_previously_parsed(self):
+        expr = "x + x"
+        nodes = self.s.parse(expr)
+        self.s.names = {"x": 21}
+        self.assertEqual(self.s.eval(expr, nodes), 42)
+
+        # This can all be done with unittest.mock.patch in python3.3+ - when we drop
+        # python2 - we can drop this nonsense.
+        class MockedCalled(Exception):
+            pass
+
+        def go_boom(*args, **kwargs):
+            raise MockedCalled("you should never see this.")
+
+        self.s.parse = go_boom
+
+        # Prove the mock is installed in self.s
+        with self.assertRaises(MockedCalled):
+            self.s.eval("10 + 10")
+
+        # Prove it's not installed in the actual SimpleEval
+        SimpleEval().eval("10 + 10")
+
+        # Now running .eval with a previously parsed
+        self.assertEqual(self.s.eval(expr, previously_parsed=nodes), 42)
+
+        self.s.names = {"x": 100}
+        self.assertEqual(self.s.eval(expr, nodes), 200)
+
 
 class TestFunctions(DRYTest):
     """Functions for expressions to play with"""
@@ -258,29 +310,29 @@ class TestFunctions(DRYTest):
         self.t("foo()", 42)
 
     def test_function_args_required(self):
-        def foo(toret):
-            return toret
+        def foo(to_return):
+            return to_return
 
         self.s.functions["foo"] = foo
         with self.assertRaises(TypeError):
             self.t("foo()", 42)
 
         self.t("foo(12)", 12)
-        self.t("foo(toret=100)", 100)
+        self.t("foo(to_return=100)", 100)
 
     def test_function_args_defaults(self):
-        def foo(toret=9999):
-            return toret
+        def foo(to_return=9999):
+            return to_return
 
         self.s.functions["foo"] = foo
         self.t("foo()", 9999)
 
         self.t("foo(12)", 12)
-        self.t("foo(toret=100)", 100)
+        self.t("foo(to_return=100)", 100)
 
     def test_function_args_bothtypes(self):
-        def foo(mult, toret=100):
-            return toret * mult
+        def foo(mult, to_return=100):
+            return to_return * mult
 
         self.s.functions["foo"] = foo
         with self.assertRaises(TypeError):
@@ -289,10 +341,10 @@ class TestFunctions(DRYTest):
         self.t("foo(2)", 200)
 
         with self.assertRaises(TypeError):
-            self.t("foo(toret=100)", 100)
+            self.t("foo(to_return=100)", 100)
 
-        self.t("foo(4, toret=4)", 16)
-        self.t("foo(mult=2, toret=4)", 8)
+        self.t("foo(4, to_return=4)", 16)
+        self.t("foo(mult=2, to_return=4)", 8)
         self.t("foo(2, 10)", 20)
 
 
@@ -328,7 +380,7 @@ class TestTryingToBreakOut(DRYTest):
         """exponent operations can take a long time."""
         old_max = simpleeval.MAX_POWER
 
-        self.t("9**9**5", 9 ** 9 ** 5)
+        self.t("9**9**5", 9**9**5)
 
         with self.assertRaises(simpleeval.NumberTooHigh):
             self.t("9**9**8", 0)
@@ -349,9 +401,15 @@ class TestTryingToBreakOut(DRYTest):
         with self.assertRaises(simpleeval.NumberTooHigh):
             self.t("1<<25000", 0)
 
+        with self.assertRaises(simpleeval.NumberTooHigh):
+            self.t("%s<<25" % (simpleeval.MAX_SHIFT_BASE + 1), 0)
+
         with self.assertRaises(simpleeval.NumberTooHigh):
             self.t("1>>25000", 0)
 
+        with self.assertRaises(simpleeval.NumberTooHigh):
+            self.t("%s>>25" % (simpleeval.MAX_SHIFT_BASE + 1), 0)
+
         # and test we can change it:
 
         old_max = simpleeval.MAX_SHIFT
@@ -420,11 +478,6 @@ class TestTryingToBreakOut(DRYTest):
         with self.assertRaises(simpleeval.IterableTooLong):
             self.t("('spam spam spam' * 5000).split() * 5000", None)
 
-    def test_python_stuff(self):
-        """other various pythony things."""
-        # it only evaluates the first statement:
-        self.t("11; x = 21; x + x", 11)
-
     def test_function_globals_breakout(self):
         """by accessing function.__globals__ or func_..."""
         # thanks perkinslr.
@@ -559,6 +612,24 @@ class TestCompoundTypes(DRYTest):
         self.t('{"a": 24}.get("b", 11)', 11)
         self.t('"a" in {"a": 24}', True)
 
+    @unittest.skipIf(not simpleeval.PYTHON35, "feature not supported")
+    def test_dict_star_expression(self):
+        self.s.names["x"] = {"a": 1, "b": 2}
+        self.t('{"a": 0, **x, "c": 3}', {"a": 1, "b": 2, "c": 3})
+
+        # and multiple star expressions should be fine too...
+        self.s.names["y"] = {"x": 1, "y": 2}
+        self.t('{"a": 0, **x, **y, "c": 3}', {"a": 1, "b": 2, "c": 3, "x": 1, "y": 2})
+
+    @unittest.skipIf(not simpleeval.PYTHON35, "feature not supported")
+    def test_dict_invalid_star_expression(self):
+        self.s.names["x"] = {"a": 1, "b": 2}
+        self.s.names["y"] = {"x": 1, "y": 2}
+        self.s.names["z"] = 42
+
+        with self.assertRaises(TypeError):
+            self.t('{"a": 0, **x, **y, **z, "c": 3}', {"a": 1, "b": 2, "c": 3})
+
     def test_tuple(self):
         self.t("()", ())
         self.t("(1,)", (1,))
@@ -589,6 +660,19 @@ class TestCompoundTypes(DRYTest):
 
         self.t('"b" in ["a","b"]', True)
 
+    @unittest.skipIf(not simpleeval.PYTHON3, "feature not supported")
+    def test_list_star_expression(self):
+        self.s.names["x"] = [1, 2, 3]
+        self.t('["a", *x, "b"]', ["a", 1, 2, 3, "b"])
+
+    @unittest.skipIf(not simpleeval.PYTHON3, "feature not supported")
+    def test_list_invalid_star_expression(self):
+        self.s.names["x"] = [1, 2, 3]
+        self.s.names["y"] = 42
+
+        with self.assertRaises(TypeError):
+            self.t('["a", *x, *y, "b"]', ["a", 1, 2, 3, "b"])
+
     def test_set(self):
         self.t("{1}", {1})
         self.t("{1, 2, 1, 2, 1, 2, 1}", {1, 2})
@@ -708,9 +792,11 @@ class TestNames(DRYTest):
 
         self.s.names["s"] = 21
 
+        # or if you attempt to assign an unknown name to another
         with self.assertRaises(NameNotDefined):
             with warnings.catch_warnings(record=True) as ws:
                 self.t("s += a", 21)
+        self.assertIsInstance(ws[0].message, simpleeval.AssignmentAttempted)
 
         self.s.names = None
 
@@ -735,6 +821,7 @@ class TestNames(DRYTest):
         # however, you can't assign to those names:
         with warnings.catch_warnings(record=True) as ws:
             self.t("a = 200", 200)
+        self.assertIsInstance(ws[0].message, simpleeval.AssignmentAttempted)
 
         self.assertEqual(self.s.names["a"], 42)
 
@@ -744,6 +831,7 @@ class TestNames(DRYTest):
 
         with warnings.catch_warnings(record=True) as ws:
             self.t("b[0] = 11", 11)
+        self.assertIsInstance(ws[0].message, simpleeval.AssignmentAttempted)
 
         self.assertEqual(self.s.names["b"], [0])
 
@@ -766,6 +854,7 @@ class TestNames(DRYTest):
 
         with warnings.catch_warnings(record=True) as ws:
             self.t("c['b'] = 99", 99)
+        self.assertIsInstance(ws[0].message, simpleeval.AssignmentAttempted)
 
         self.assertFalse("b" in self.s.names["c"])
 
@@ -775,6 +864,7 @@ class TestNames(DRYTest):
 
         with warnings.catch_warnings(record=True) as ws:
             self.t("c['c']['c'] = 21", 21)
+        self.assertIsInstance(ws[0].message, simpleeval.AssignmentAttempted)
 
         self.assertEqual(self.s.names["c"]["c"]["c"], 11)
 
@@ -789,6 +879,7 @@ class TestNames(DRYTest):
 
         with warnings.catch_warnings(record=True) as ws:
             self.t("a.b.c = 11", 11)
+        self.assertIsInstance(ws[0].message, simpleeval.AssignmentAttempted)
 
         self.assertEqual(self.s.names["a"]["b"]["c"], 42)
 
@@ -816,6 +907,7 @@ class TestNames(DRYTest):
 
     def test_object(self):
         """using an object for name lookup"""
+        # pylint: disable=attribute-defined-outside-init
 
         class TestObject(object):
             @staticmethod
@@ -1032,10 +1124,12 @@ class TestUnusualComparisons(DRYTest):
                 return BinaryExpression("LT")
 
         b = Blah()
-        self.s.names = {"b": b}
-        # This should not crash:
-        e = eval("b > 2", self.s.names)
+        # These should not crash:
+        self.assertEqual(b > 2, BinaryExpression("GT"))
+        self.assertEqual(b < 2, BinaryExpression("LT"))
 
+        # And should also work in simpleeval
+        self.s.names = {"b": b}
         self.t("b > 2", BinaryExpression("GT"))
         self.t("1 < 5 > b", BinaryExpression("LT"))
 
@@ -1056,7 +1150,7 @@ class TestGetItemUnhappy(DRYTest):
 
         self.assertEqual(m.anything, 42)
         with self.assertRaises(NotImplementedError):
-            m["nothing"]
+            m["nothing"]  # pylint: disable=pointless-statement
 
         self.s.names = {"m": m}
         self.t("m.anything", 42)
@@ -1108,6 +1202,7 @@ class TestDisallowedFunctions(DRYTest):
     def test_functions_are_disallowed_at_init(self):
         DISALLOWED = [type, isinstance, eval, getattr, setattr, help, repr, compile, open]
         if simpleeval.PYTHON3:
+            # pylint: disable=exec-used
             exec("DISALLOWED.append(exec)")  # exec is not a function in Python2...
 
         for f in simpleeval.DISALLOW_FUNCTIONS:
@@ -1115,12 +1210,13 @@ class TestDisallowedFunctions(DRYTest):
 
         for x in DISALLOWED:
             with self.assertRaises(FeatureNotAvailable):
-                s = SimpleEval(functions={"foo": x})
+                SimpleEval(functions={"foo": x})
 
     def test_functions_are_disallowed_in_expressions(self):
         DISALLOWED = [type, isinstance, eval, getattr, setattr, help, repr, compile, open]
 
         if simpleeval.PYTHON3:
+            # pylint: disable=exec-used
             exec("DISALLOWED.append(exec)")  # exec is not a function in Python2...
 
         for f in simpleeval.DISALLOW_FUNCTIONS:
@@ -1138,11 +1234,13 @@ class TestDisallowedFunctions(DRYTest):
         simpleeval.DEFAULT_FUNCTIONS = DF.copy()
 
 
-@unittest.skipIf(simpleeval.PYTHON3 != True, "Python2 fails - but it's not supported anyway.")
+@unittest.skipIf(simpleeval.PYTHON3 is not True, "Python2 fails - but it's not supported anyway.")
 @unittest.skipIf(platform.python_implementation() == "PyPy", "GC set_debug not available in PyPy")
 class TestReferenceCleanup(DRYTest):
     """Test cleanup without cyclic references"""
 
+    # pylint: disable=attribute-defined-outside-init
+
     def setUp(self):
         self._initial_gc_isenabled = gc.isenabled()
 
@@ -1165,5 +1263,35 @@ class TestReferenceCleanup(DRYTest):
         simpleeval.SimpleEval()
 
 
+class TestNoEntries(DRYTest):
+    def test_no_functions(self):
+        self.s.eval("int(42)")
+        with self.assertRaises(FunctionNotDefined):
+            s = SimpleEval(functions={})
+            s.eval("int(42)")
+
+    def test_no_names(self):
+        # does not work on current Py3, True et al. are keywords now
+        self.s.eval("True")
+        # with self.assertRaises(NameNotDefined):
+        s = SimpleEval(names={})
+        if sys.version_info < (3,):
+            with self.assertRaises(NameNotDefined):
+                s.eval("True")
+        else:
+            s.eval("True")
+
+    def test_no_operators(self):
+        self.s.eval("1+2")
+        self.s.eval("~2")
+        s = SimpleEval(operators={})
+
+        with self.assertRaises(OperatorNotDefined):
+            s.eval("1+2")
+
+        with self.assertRaises(OperatorNotDefined):
+            s.eval("~ 2")
+
+
 if __name__ == "__main__":  # pragma: no cover
     unittest.main()
